from __future__ import (division, print_function, unicode_literals,
                        absolute_import)
import logging
import os

from swingers.models import Audit
from swingers import models
from swingers.utils.decorators import workdir

from django.utils.encoding import python_2_unicode_compatible
from django.utils import timezone
from django.db.models import Max, Q
from smart_selects.db_fields import ChainedForeignKey

from pbs.document.fields import ContentTypeRestrictedFileField
from pbs.document.utils import get_dimensions
from pbs.prescription.models import Prescription

logger = logging.getLogger(__name__)


def content_file_name(self, filename):
    if filename.rsplit('.')[-1] == "zip":
        extension = "zip"
    else:
        extension = "pdf"
    return "uploads/{0}/{0}_{1}_{2}_{3}.{4}".format(
        str(self.prescription.season).strip().replace(" ", "_"),
        self.prescription.burn_id,
        self.descriptor.strip().replace(" ", "_"),
        timezone.localtime(timezone.now()).isoformat(
        ).rsplit(".")[0].replace(":", "")[:-2],
        extension)


@python_2_unicode_compatible
class DocumentCategory(models.Model):
    name = models.CharField(max_length=200)
    order = models.PositiveSmallIntegerField(default=0)

    class Meta:
        verbose_name = "Document Category"
        verbose_name_plural = "Document Categories"
        ordering = ['name']

    def __str__(self):
        return self.name


class CategoryManager(models.Manager):
    use_for_related_fields = True

    def _query_by_names(self, *names):
        return reduce(lambda x, y: x | y,
                      (Q(name__iexact=name) for name in names),
                      Q())

    def not_tag_names(self, *names):
        return self.get_query_set().exclude(self._query_by_names(*names))

    def tag_names(self, *names):
        return self.get_query_set().filter(self._query_by_names(*names))


@python_2_unicode_compatible
class DocumentTag(models.Model):
    name = models.CharField(verbose_name="Document Tag", max_length=200)
    category = models.ForeignKey(DocumentCategory)
    objects = CategoryManager()

    class Meta:
        ordering = ['name']

    def __str__(self):
        return self.name


class TagManager(models.Manager):
    use_for_related_fields = True

    def partd(self):
        qs = self.not_tag_names("Traffic Diagrams",
                                "Sign Inspection and Surveillance Form",
                                ("Prescribed Burning Organisational Structure "
                                 "and Communications Plan"),
                                "Context Map",
                                "Prescribed Burning SMEAC Checklist")
        qs.modified = qs.aggregate(Max('modified'))["modified__max"]
        return qs

    def _query_by_names(self, *names):
        return reduce(lambda x, y: x | y,
                      (Q(tag__name__iexact=name) for name in names),
                      Q())

    def not_tag_names(self, *names):
        return self.get_query_set().exclude(self._query_by_names(*names))

    def tag_names(self, *names):
        return self.get_query_set().filter(self._query_by_names(*names))

    def __getattr__(self, name):
        if name[:4] == "tag_":
            qs = self.get_query_set().filter(
                tag__name__iexact=name[4:].replace("_", " "))
            qs.modified = qs.aggregate(Max('modified'))["modified__max"]
            return qs
        else:
            return super(TagManager, self).__getattr__(name)

    def printable(self):
        """
        Return the set of printable documents. This removes any zip files from
        the set of documents.
        """
        qs = self.get_query_set()
        return filter(lambda x: x.filename.endswith('.pdf'), qs)


@python_2_unicode_compatible
class Document(Audit):
    prescription = models.ForeignKey(
        Prescription, null=True,
        help_text="Prescription that this document belongs to")
    category = models.ForeignKey(DocumentCategory, related_name="documents")
    tag = ChainedForeignKey(
        DocumentTag, chained_field="category", chained_model_field="category",
        show_all=False, auto_choose=True, verbose_name="Descriptor")
    custom_tag = models.CharField(
        max_length=64, blank=True, verbose_name="Custom Descriptor")
    document = ContentTypeRestrictedFileField(
        upload_to=content_file_name, max_length=200,
        content_types=['application/pdf', 'image/tiff', 'image/tif',
                       'image/jpeg', 'image/jpg', 'image/gif', 'image/png',
                       'application/zip', 'application/x-zip-compressed'],
        help_text='Acceptable file types: pdf, tiff, jpg, gif, png, zip')

    objects = TagManager()

    @property
    def descriptor(self):
        if self.custom_tag:
            return "Other ({0})".format(self.custom_tag)
        else:
            return self.tag.name

    @property
    def dimensions(self):
        return get_dimensions(self.document.path)

    @property
    def filename(self):
        return os.path.basename(self.document.path)

    class Meta:
        ordering = ['tag', 'document']

    def __str__(self):
        return "{0} - {1}".format(self.prescription, self.document.name)
