from __future__ import unicode_literals, absolute_import

import os
import subprocess
import json
import unicodecsv

from functools import update_wrapper, partial
from dateutil import tz
from datetime import datetime

from swingers.admin import DetailAdmin

from guardian.shortcuts import assign_perm

from django.conf import settings
from django.contrib import admin
from django.contrib import messages
from django.contrib.admin import helpers
from django.contrib.admin.util import quote, unquote, flatten_fieldsets
from django.contrib.auth.models import Group
from django.contrib.auth.decorators import permission_required
from django.core.exceptions import (FieldError, ValidationError,
                                    PermissionDenied)
from django.core.urlresolvers import reverse
from django.db import transaction, router
from django.http import HttpResponse, HttpResponseRedirect, Http404
from django.forms.models import modelform_factory
from django.template import RequestContext
from django.template.defaultfilters import date
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.templatetags.tz import localtime
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.utils.encoding import force_text
from django.utils.html import escape
from django.utils.translation import ugettext as _, ugettext_lazy
from django.views.decorators.csrf import csrf_protect

from pbs.admin import BaseAdmin, get_permission_codename
from pbs.prescription.actions import (delete_selected,
                                      delete_approval_endorsement)
from pbs.prescription.forms import (PrescriptionCreateForm, EndorsingRoleForm,
                                    AddEndorsementForm, AddApprovalForm,
                                    PrescriptionEditForm,
                                    PrescriptionPriorityForm,
                                    BriefingChecklistForm,
                                    PrescriptionSummaryForm)
from pbs.prescription.models import (Season, Prescription, RegionalObjective,
                                     Objective, Region)

from pbs.report.models import Evaluation
from pbs.report.forms import (SummaryCompletionStateForm,
                              BurnImplementationStateForm,
                              BurnClosureStateForm)

from pbs.templatetags.pbs_markdown import markdownify
from pbs.utils import get_deleted_objects, update_permissions
from pbs.utils.widgets import CheckboxSelectMultiple

csrf_protect_m = method_decorator(csrf_protect)


class DistrictAdmin(admin.ModelAdmin):
    list_display = ('name', 'region', 'code')


class PrescriptionAdmin(DetailAdmin, BaseAdmin):
    """
    An admin class to manage prescriptions.
    Caution: self.fieldsets and self.form are both declared as the fieldsets
    and form for the *add* form. The functions `get_form` and `get_fieldsets`
    have been overridden to return a different form and fieldsets based on the
    selected prescription's status.
    """
    objectives_template = 'admin/prescription/prescription/add_objectives.html'
    summary_template = 'admin/prescription/prescription/summary.html'
    pre_summary_template = 'admin/prescription/prescription/pre_summary.html'
    day_summary_template = 'admin/prescription/prescription/day_summary.html'
    post_summary_template = 'admin/prescription/prescription/post_summary.html'
    corporate_approval_template = 'admin/prescription/prescription/cbas.html'
    remove_selected_confirmation_template = (
        'admin/prescription/prescription/remove_selected_confirmation.html')

    changelist_link_detail = True
    list_display = ("burn_id", "name", "region", "district",
                    "planned_year", "planned_season",
                    "contentious", "aircraft_burn", "allocation",
                    "priority", "remote_sensing_priority",
                    "planned_year", "planned_season",
                    "contentious", "aircraft_burn", "allocation",
                    "priority", "remote_sensing_priority",
                    "planning_status", "endorsement_status", "approval_status",
                    "approval_expiry", "ignition_status", "status",
                    "prescribing_officer_name", "date_modified_local")

    actions = [delete_selected, 'export_to_csv', delete_approval_endorsement]

    form = PrescriptionCreateForm
    edit_form = PrescriptionEditForm

    fieldsets = (
        (None, {
            "fields": ('name', ('planned_season', 'planned_year'),
                       ('last_season', 'last_year'),
                       ('last_year_unknown', 'last_season_unknown'),
                       ('region', 'district'),
                       'location', 'forest_blocks',
                       ('contentious', 'contentious_rationale'),
                       ('aircraft_burn', 'remote_sensing_priority'),
                       'purposes',
                       'allocation', 'treatment_percentage',
                       'area', 'perimeter')
        }),
    )
    search_fields = ('name', 'burn_id', 'location')
    filter_fields = (
        ('district', {'filter_on': 'region'}),
        ('shire', {'filter_on': 'district'}),
    )
    filter_horizontal = ('tenures', 'vegetation_types', 'shires',
                         'forecast_areas', 'endorsing_roles')

    def prescribing_officer_name(self, obj):
        if obj.prescribing_officer is not None:
            return obj.prescribing_officer.get_full_name()
        else:
            return "No prescribing officer"
    prescribing_officer_name.admin_order_field = 'prescribing_officer'
    prescribing_officer_name.short_description = 'Prescribing Officer'

    def date_modified_local(self, obj):
        return obj.date_modified_local()
    date_modified_local.admin_order_field = 'modified'
    date_modified_local.short_description = 'Date Modified'

    def approval_expiry(self, obj):
        if obj.is_approved:
            return date(obj.current_approval.valid_to)
        else:
            return "Not applicable"
    approval_expiry.short_description = "Approved Until"

    def get_urls(self):
        """
        Add some extra views for handling the prescription summaries and a page
        to handle selecting RFMP objectives for a burn.
        """
        from django.conf.urls import patterns, url

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.module_name

        urlpatterns = patterns(
            '',
            url(r'^(\d+)/add/objectives/$',
                wrap(self.add_objectives),
                name='%s_%s_objectives' % info),
            url(r'^(\d+)/summary/$',
                wrap(self.summary),
                name='%s_%s_summary' % info),
            url(r'^(\d+)/summary/pre/$',
                wrap(self.pre_summary),
                name='%s_%s_pre_summary' % info),
            url(r'^(\d+)/summary/day/$',
                wrap(self.day_summary),
                name='%s_%s_day_summary' % info),
            url(r'^(\d+)/summary/post/$',
                wrap(self.post_summary),
                name='%s_%s_post_summary' % info),
            url(r'^(\d+)/download/$',
                wrap(self.lualatex),
                name='%s_%s_download' % info),
            url(r'^(\d+)/export/$',
                wrap(self.lualatex),
                name='%s_%s_export' % info),
            url(r'^(\d+)/cbas/$',
                wrap(self.corporate_approve),
                name='%s_%s_corporate_approve' % info),
            url(r'^(\d+)/endorsement/$',
                wrap(self.endorse),
                name='%s_%s_endorse' % info),
            url(r'^(\d+)/endorsement/officers$',
                wrap(self.endorsing_roles),
                name='%s_%s_endorsing_roles' % info),
            url(r'^(\d+)/approval/$',
                wrap(self.approve),
                name='%s_%s_approve' % info),
            url(r'^(\d+)/closure/$',
                wrap(self.close),
                name='%s_%s_close' % info),
            url(r'^(\d+)/sitemap/$',
                wrap(self.sitemap),
                name='%s_%s_sitemap' % info),
        )

        return urlpatterns + super(PrescriptionAdmin, self).get_urls()

    def changelist_view(self, request, extra_context=None):
        # figure out how to auto-select current user's region from
        # request.user.profile
        context = {
            'regions': Region.objects.all(),
            'seasons': Season.SEASON_CHOICES,
        }
        context.update(extra_context or {})
        return super(PrescriptionAdmin, self).changelist_view(
            request, extra_context=context)

    def get_actions(self, request):
        actions = super(PrescriptionAdmin, self).get_actions(request)
        if not request.user.has_perm('prescription.delete_prescription'):
            if actions['delete_selected']:
                del actions['delete_selected']
        if not request.user.has_perm('prescription.can_delete_approval'):
            if actions['delete_approval_endorsement']:
                del actions['delete_approval_endorsement']

        return actions

    def export_to_csv(self, request, queryset):
        # TODO: fix up the date/time formatting to use the default template
        # filters.
        local_zone = tz.tzlocal()
        response = HttpResponse(content_type="text/csv")
        response['Content-Disposition'] = "attachment; filename=export.csv"

        writer = unicodecsv.writer(response, quoting=unicodecsv.QUOTE_ALL)
        writer.writerow([
            'Region', 'District', 'Name of Burn', 'Burn ID', 'Date Updated',
            'Planning Status', 'Endorsement Status', 'Approval Status',
            'Ignition Status', 'Burn Closed', 'Burn Complexity',
            'Burn Priority', 'Burn Risk', 'Contentious',
            'Location of Burn', 'Prescribing Officer', 'Year', 'Season',
            'Proposed Ignition Type', 'Actual Ignition Type',
            'Fire Weather Forecast Area/s', 'Treatment %', 'Planned Burn Area',
            'Planned Burn Perimeter', 'Treatment Area',
            'Total Burnt Area Estimate', 'Length of Edging/Perimeter',
            'Depth of Edging', 'Shire', 'Bushfire Act Zone',
            'Prohibited Period', 'Burn Purpose/s', 'Burn Objective/s',
            'Ignition Date', 'Program', 'Vegetation Type',
            'Fuel Type', 'Land Tenure', 'Year Last Burnt', 'Season Last Burnt',
            'Date(s) Escaped', 'DPaW Fire Number', 'DFES Fire Number',
            'Shortcode', 'Remote Sensing Priority', 'Aircraft Burn',
            'Success Criteria', 'Success Criteria Achieved',
            'Observations Identified', 'Proposed Action',
            'Endorsement Name/s', 'Endorsement Date',
            'Approval Name/s', 'Approval Date', 'Approved Until'])

        for item in queryset:
            if item.last_season is not None:
                last_season_burnt = item.get_last_season_display()
            else:
                last_season_burnt = "Unknown"
            if item.last_year is not None:
                last_year_burnt = item.last_year
            else:
                last_year_burnt = "Unknown"

            if item.is_approved:
                approved_until = date(item.current_approval.valid_to)
            else:
                approved_until = ""

            proposed_ignitions = ('"%s"' % ", ".join(
                ",".join('{0}:{1}'.format(x.seqno, ignition_type)
                         for ignition_type
                         in (x.ignition_types.all() or ("Unset",))
                         )
                for x in item.lightingsequence_set.order_by('seqno'))
            )
            actual_ignitions = ('"%s"' % ",".join(
                ",".join('{0}:{1}'.format(x.ignition.strftime('%d/%m/%Y'),
                                          ignition_type)
                         for ignition_type
                         in (x.ignition_types.all() or ("Unset",))
                         )
                for x in item.areaachievement_set.order_by('ignition'))
            )
            forecast_areas = '"%s"' % ",".join(x.name for x
                                               in item.forecast_areas.all())
            treatment_area = item.total_treatment_area
            burnt_area_estimate = item.total_burnt_area_estimate
            length_of_edging = item.total_edging_length
            depth_of_edging = item.total_edging_depth
            objectives = '"%s"' % ",".join(x.objectives for x in item.objective_set.all())
            shires = '"%s"' % ",".join(x.name for x in item.shires.all())
            purposes = '"%s"' % ",".join(x.name for x in item.purposes.all())
            veg_types = '"%s"' % ",".join(x.name for x in item.vegetation_types.all())
            fuel_descriptions = '"%s"' % ",".join(x.fuel_description for x in item.lightingsequence_set.all())
            tenures = '"%s"' % ",".join(x.name for x in item.tenures.all())
            escape_dates = ", ".join([datetime.strftime(x.date_escaped, '%d/%m/%Y') for x in item.areaachievement_set.filter(date_escaped__isnull=False).order_by('ignition')])
            dpaw_fire_nums = '"%s"' % ",".join(x.dpaw_fire_no for x in item.areaachievement_set.exclude(dpaw_fire_no__isnull=True).exclude(dpaw_fire_no__exact='').order_by('ignition'))
            dfes_fire_nums = '"%s"' % ",".join(x.dfes_fire_no for x in item.areaachievement_set.exclude(dfes_fire_no__isnull=True).exclude(dfes_fire_no__exact='').order_by('ignition'))
            success_criterias = '"%s"' % ",".join(x.criteria for x in item.successcriteria_set.all().order_by('id'))
            success_criteria_outcomes = '"%s"' % ",".join(x.criteria.criteria + ':' + x.get_achieved_display() + ':' + x.summary for x in Evaluation.objects.filter(criteria__in=item.successcriteria_set.all().order_by('id')).exclude(achieved__isnull=True).exclude(summary__isnull=True))
            observations = '"%s"' % ",".join((x.observations or "No observation") + ':' + (x.action or "No action") for x in item.proposedaction_set.all().order_by('id'))
            proposed_actions = '"%s"' % ",".join((x.observations or "No observation") + ':' + (x.action or "No action") for x in item.proposedaction_set.all().order_by('id'))
            endorsements = '"%s"' % ",".join(x.role.name + ':' + x.get_endorsed_display() for x in item.endorsement_set.all())
            approvals = '"%s"' % ",".join(x.creator.get_full_name() + ':' + x.valid_to.strftime('%d/%m/%Y') + '(%s)' % x.extension_count for x in item.approval_set.all())

            writer.writerow([
                item.region, item.district, item.name, item.burn_id,
                item.modified.astimezone(local_zone).strftime('%d/%m/%Y %H:%M:%S'),
                item.get_planning_status_display(),
                item.get_endorsement_status_display(),
                item.get_approval_status_display(),
                item.get_ignition_status_display(),
                "Yes" if item.status == item.STATUS_CLOSED else "No",
                item.maximum_complexity,
                item.get_priority_display(), item.maximum_risk,
                "Yes" if item.contentious else "No", item.location,
                item.prescribing_officer, item.planned_year,
                item.get_planned_season_display(),
                proposed_ignitions, actual_ignitions, forecast_areas,
                item.treatment_percentage, item.area, item.perimeter,
                treatment_area, burnt_area_estimate,
                length_of_edging, depth_of_edging,
                shires, item.bushfire_act_zone, item.prohibited_period,
                purposes, objectives,
                actual_ignitions,
                item.get_allocation_display(), veg_types, fuel_descriptions,
                tenures, last_year_burnt, last_season_burnt, escape_dates,
                dpaw_fire_nums, dfes_fire_nums, item.short_code,
                item.get_remote_sensing_priority_display(),
                "Yes" if item.aircraft_burn else "No", success_criterias,
                success_criteria_outcomes, observations, proposed_actions,
                endorsements,
                item.endorsement_status_modified.astimezone(
                    local_zone).strftime('%d/%m/%Y %H:%M:%S') if item.endorsement_status_modified else "",
                approvals,
                item.approval_status_modified.astimezone(
                    local_zone).strftime('%d/%m/%Y %H:%M:%S') if item.approval_status_modified else "",
                approved_until
            ])

        return response
    export_to_csv.short_description = ugettext_lazy("Export to CSV")

    def response_post_save_add(self, request, obj):
        """
        Override the redirect url after successful save of a new burn plan.
        """
        url = reverse('admin:prescription_prescription_detail',
                      args=[str(obj.id)])
        return HttpResponseRedirect(url)

    def response_post_save_change(self, request, obj):
        """
        Override the redirect url after successful save of an existing burn
        plan.
        """
        url = reverse('admin:prescription_prescription_detail',
                      args=[str(obj.id)])
        return HttpResponseRedirect(url)

    def get_fieldsets(self, request, obj=None):
        """
        Tweak fieldsets based on whether the user is creating a new
        prescription or editing an existing one.
        """
        if obj:
            return (('Corporate Burn Attribute Summary', {
                "fields": ('name', ('planned_year', 'planned_season'),
                           ('last_year', 'last_season',
                            'last_year_unknown', 'last_season_unknown'),
                           'region', 'district', 'location',
                           'forest_blocks',
                           ('contentious', 'contentious_rationale'),
                           'aircraft_burn', 'allocation',
                           'remote_sensing_priority', 'purposes',
                           ('area', 'perimeter', 'treatment_percentage'))
            }), ('Other Attributes', {
                "fields": ('tenures', 'vegetation_types', 'shires',
                           'bushfire_act_zone', 'prohibited_period',
                           'forecast_areas', 'prescribing_officer',
                           'short_code')
            }))
        else:
            return super(PrescriptionAdmin, self).get_fieldsets(request, obj)

    def get_readonly_fields(self, request, obj=None):
        if obj and obj.planning_status != obj.PLANNING_DRAFT:
            return ('name', 'burn_id', 'planned_year', 'planned_season',
                    'last_year', 'last_season', 'last_season_unknown',
                    'last_year_unknown',
                    'region', 'district', 'contentious',
                    'contentious_rationale')
        elif obj and not obj.is_draft:
            return (None,)
        else:
            return super(PrescriptionAdmin, self).get_readonly_fields(
                request, obj)

    def formfield_for_dbfield(self, db_field, **kwargs):

        today = timezone.now().date()
        if db_field.name == 'planned_year':
            kwargs['initial'] = today.year

        # This is commented out because there are two possible seasons
        # for each period. We need a dataset to relate the region to the
        # seasons they will use (tropical, temperate), otherwise we will
        # not be able to pre-populate this field.
        #season = Season.objects.get(start__lte=today, end__gte=today)
        #if db_field.name == 'planned_season':
        #    kwargs['initial'] = season.name

        return super(PrescriptionAdmin, self).formfield_for_dbfield(
            db_field, **kwargs)

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """
        Populate some of the foreign keys with initial data.
        """
        profile = request.user.get_profile()
        if db_field.name == 'region' and profile.region is not None:
            kwargs['initial'] = profile.region.pk
            return db_field.formfield(**kwargs)

        if db_field.name == 'district' and profile.district is not None:
            kwargs['initial'] = profile.district.pk
            return db_field.formfield(**kwargs)

        if db_field.name == 'prescribing_officer':
            kwargs['initial'] = request.user.pk
            field = db_field.formfield(**kwargs)
            return field

        return super(PrescriptionAdmin, self).formfield_for_foreignkey(
            db_field, request, **kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """
        Replace the widget for the burn purposes with our own checkbox select.
        """
        if db_field.name == 'purposes':
            kwargs['widget'] = CheckboxSelectMultiple()

        return super(PrescriptionAdmin, self).formfield_for_manytomany(
            db_field, request, **kwargs)

    def get_form_class(self, request, obj=None):
        if obj:
            return self.edit_form
        else:
            return self.form

    def get_form(self, request, obj=None, **kwargs):
        """
        Returns a Form class for use in the admin add view. This is used by
        add_view and change_view.
        """
        if self.declared_fieldsets:
            fields = flatten_fieldsets(self.get_fieldsets(request, obj))
        else:
            fields = None
        if self.exclude is None:
            exclude = []
        else:
            exclude = list(self.exclude)
        exclude.extend(self.get_readonly_fields(request, obj))

        form = self.get_form_class(request, obj)

        # a simple hack to set the default prescribing officer
        if obj is not None and obj.prescribing_officer is None:
            obj.prescribing_officer = request.user

        if self.exclude is None and (
            hasattr(form, '_meta') and form._meta.exclude
        ):
            # Take the custom ModelForm's Meta.exclude into account only if the
            # ModelAdmin doesn't define its own.
            exclude.extend(self.form._meta.exclude)
        # if exclude is an empty list we pass None to be consistent with the
        # default on modelform_factory
        exclude = exclude or None
        defaults = {
            "form": form,
            "fields": fields,
            "exclude": exclude,
            "formfield_callback": partial(self.formfield_for_dbfield,
                                          request=request),
        }
        defaults.update(kwargs)

        try:
            return modelform_factory(self.model, **defaults)
        except FieldError as e:
            raise FieldError('%s. Check fields/fieldsets/exclude'
                             ' attributes of class %s.'
                             % (e, self.__class__.__name__))

    def detail_view(self, request, object_id, extra_context=None):
        obj = self.get_object(request, unquote(object_id))

        # the epfp approval modal
        submit_for_epfp_approval_title = self._approve_title(obj)
        AdminAddApprovalForm = self._approve_approval_form(request)
        form = AdminAddApprovalForm(initial={'prescription': obj})
        submit_for_epfp_approval_form, media = self._approve_form(
            request, obj, form)

        context = {
            'current': obj,
            'submit_for_epfp_approval_form': submit_for_epfp_approval_form,
            'submit_for_epfp_approval_title': submit_for_epfp_approval_title,
            'media': media,
        }
        context.update(extra_context or {})

        return super(PrescriptionAdmin, self).detail_view(
            request, object_id, context)

    def corporate_approve(self, request, object_id, extra_context=None):
        """
        View to manage corporate approval of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))
        if request.method == 'POST':
            url = reverse('admin:prescription_prescription_detail',
                          args=[str(obj.id)])
            if request.POST.get('_cancel'):
                return HttpResponseRedirect(url)
            if ((obj.planning_status == obj.PLANNING_DRAFT
                 and obj.can_corporate_approve)):
                obj.planning_status = obj.PLANNING_SUBMITTED
                obj.planning_status_modified = timezone.now()
                obj.save()
                self.message_user(
                    request, "Successfully submitted for corporate approval.")
                return HttpResponseRedirect(url)
            if obj.planning_status == obj.PLANNING_SUBMITTED:
                # Only FMSB data coordinators can apply corporate approval
                if ((not request.user.has_perm(
                     'prescription.can_corporate_approve'))):
                    raise PermissionDenied

                obj.planning_status = obj.PLANNING_APPROVED
                obj.planning_status_modified = timezone.now()
                obj.save()
                self.message_user(
                    request, "Corporate approval successful.")
                return HttpResponseRedirect(url)

        context = {
            'current': obj,
        }
        return TemplateResponse(request, self.corporate_approval_template,
                                context, current_app=self.admin_site.name)

    def endorse(self, request, object_id, extra_context=None):
        """
        View to manage endorsement of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))

        if obj.endorsement_status == obj.ENDORSEMENT_DRAFT:
            title = "Submit for endorsement"
        else:
            title = "Endorse this ePFP"

        form = AddEndorsementForm(request.POST or None, request=request)

        if request.method == 'POST':
            url = reverse('admin:prescription_prescription_detail',
                          args=[str(obj.id)])
            if request.POST.get('_cancel'):
                return HttpResponseRedirect(url)
            if ((obj.endorsement_status == obj.ENDORSEMENT_DRAFT
                 and obj.can_endorse)):
                obj.endorsement_status = obj.ENDORSEMENT_SUBMITTED
                obj.endorsement_status_modified = timezone.now()
                obj.save()

                # Remove all object permissions for sections that can't be
                # modified after submitting for endorsement
                update_permissions(obj, self.admin_site, 'endorsement')

                self.message_user(
                    request, "Successfully submitted for endorsement.")
                return HttpResponseRedirect(url)
            if obj.endorsement_status == obj.ENDORSEMENT_SUBMITTED:
                # create an endorsement
                if form.is_valid():
                    endorsement = form.save(commit=False)
                    endorsement.prescription = obj
                    endorsement.endorsed = (
                        request.POST.get('_dont_endorse') is None)
                    endorsement.save()

                # check that all needed endorsements are there
                if obj.all_reviewed:
                    if obj.all_endorsed:
                        obj.endorsement_status = obj.ENDORSEMENT_APPROVED
                        msg = (" All endorsements have been completed on "
                               "this ePFP.")
                    else:
                        obj.endorsement_status = obj.ENDORSEMENT_DRAFT
                        obj.endorsement_set.all().delete()
                        msg = (" All endorsements have been reviewed but some "
                               "of them have not been endorsed on this ePFP.")
                    obj.save()
                    self.message_user(
                        request, "Successfully added endorsement." +
                        msg)
                else:
                    self.message_user(
                        request, "Successfully added endorsement.")
                return HttpResponseRedirect(url)

        form.fields['role'].queryset = obj.not_endorsed_endorsing_roles

        context = {
            'title': title,
            'current': obj,
            'errors': None,
            'form': form,
            'endorsed_endorsing_roles': obj.endorsement_set.all(),
            'not_endorsed_endorsing_roles': obj.not_endorsed_endorsing_roles,
        }
        return TemplateResponse(request, "admin/prescription/prescription/"
                                "endorsement.html", context,
                                current_app=self.admin_site.name)

    def endorsing_roles(self, request, object_id, extra_context=None):
        """
        View to manage determining additional endorsement roles in an ePFP.
        """
        class AdminEndorsingRoleForm(EndorsingRoleForm):
            formfield_callback = partial(
                self.formfield_for_dbfield, request=request)

        obj = self.get_object(request, unquote(object_id))

        if obj is None:
            raise Http404(_('%(name)s object with primary key (%key)r'
                            ' does not exist.') %
                          {'name': force_text(self.opts.verbose_name),
                           'key': object_id})

        if request.method == 'POST':
            form = AdminEndorsingRoleForm(request.POST, instance=obj)
            url = reverse('admin:prescription_prescription_detail',
                          args=[str(obj.id)])
            if request.POST.get('_cancel'):
                return HttpResponseRedirect(url)
            if form.is_valid():
                obj = form.save(commit=False)
                obj.endorsing_roles_determined = True
                obj.save()
                form.save_m2m()
                self.message_user(
                    request, "Successfully saved endorsing officers.")
                return HttpResponseRedirect(url)
        else:
            form = AdminEndorsingRoleForm(instance=obj)

        admin_form = helpers.AdminForm(
            form, [(None, {'fields': list(form.base_fields)})],
            self.get_prepopulated_fields(request, obj),
            self.get_readonly_fields(request, obj),
            model_admin=self)
        media = self.media + admin_form.media

        context = {
            'title': 'Determine endorsing roles',
            'form': admin_form,
            'media': media,
            'current': obj,
            'errors': None
        }
        return TemplateResponse(request, "admin/prescription/prescription/"
                                "endorsing_roles.html", context,
                                current_app=self.admin_site.name)

    def _approve_title(self, obj):
        if obj.approval_status == obj.APPROVAL_DRAFT:
            title = "Submit for approval"
        elif obj.approval_status == obj.APPROVAL_SUBMITTED:
            title = "Approve ePFP {0} ({1} {2})".format(
                obj.burn_id, obj.get_planned_season_display(),
                obj.planned_year)
        elif obj.approval_status == obj.APPROVAL_APPROVED:
            title = "Approve extension of ePFP {0} ({1} {2})".format(
                obj.burn_id, obj.get_planned_season_display(),
                obj.planned_year)

        return title

    def _approve_form(self, request, obj, form):
        admin_form = helpers.AdminForm(
            form, [(None, {'fields': list(form.base_fields)})],
            self.get_prepopulated_fields(request, obj),
            self.get_readonly_fields(request, obj),
            model_admin=self)
        media = self.media + admin_form.media

        return admin_form, media

    def _approve_approval_form(self, request):
        class AdminAddApprovalForm(AddApprovalForm):
            formfield_callback = partial(
                self.formfield_for_dbfield, request=request)
        return AdminAddApprovalForm

    def approve(self, request, object_id, extra_context=None):
        """
        View to manage corporate approval of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))
        title = self._approve_title(obj)

        AdminAddApprovalForm = self._approve_approval_form(request)

        form = AdminAddApprovalForm(initial={'prescription': obj})
        if request.method == 'POST':
            url = reverse('admin:prescription_prescription_detail',
                          args=[str(obj.id)])
            if obj.approval_status == obj.APPROVAL_DRAFT and obj.can_approve:
                # create an approval
                obj.approval_status = obj.APPROVAL_SUBMITTED
                obj.approval_status_modified = timezone.now()
                obj.save()
                self.message_user(
                    request, "Successfully submitted for approval.")
                return HttpResponseRedirect(url)
            elif obj.approval_status == obj.APPROVAL_SUBMITTED:
                if request.POST.get('_cancel'):
                    obj.clear_approvals()
                    self.message_user(
                        request, "Approval rejected. ePFP is now draft.")
                    return HttpResponseRedirect(url)

                form = AdminAddApprovalForm(request.POST,
                                            initial={'prescription': obj})
                if form.is_valid():
                    approval = form.save(commit=False)
                    approval.prescription = obj
                    approval.save()
                    obj.approval_status = obj.APPROVAL_APPROVED
                    obj.approval_status_modified = timezone.now()
                    obj.save()
                    self.message_user(
                        request, "Successfully approved.")
                    return HttpResponseRedirect(url)
            elif obj.approval_status == obj.APPROVAL_APPROVED:
                if request.POST.get('_cancel'):
                    self.message_user(
                        request, "Didn't extend approval.")
                    return HttpResponseRedirect(url)
                else:
                    approval = obj.current_approval
                    if approval and approval.extension_count < 3:
                        approval.extension_count = approval.extension_count + 1
                        approval.valid_to = approval.next_valid_to
                        approval.save()
                        self.message_user(
                            request, "Successfully extended approval.")
                    else:
                        self.message_user(request, "You can't extend an "
                                          "approval more than 3 times.")
                    return HttpResponseRedirect(url)

        admin_form, media = self._approve_form(request, obj, form)

        context = {
            'title': title,
            'current': obj,
            'form': admin_form,
            'media': media,
            'errors': None,
        }
        return TemplateResponse(request, "admin/prescription/prescription/"
                                "approval.html", context,
                                current_app=self.admin_site.name)

    def sitemap(self, request, object_id, extra_context=None):
        """
        View to manage closure of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))
        title = "Sitemap"

        context = {
            'title': title,
            'current': obj,
        }
        return TemplateResponse(request, "admin/prescription/prescription/"
                                "sitemap.html", context,
                                current_app=self.admin_site.name)

    def close(self, request, object_id, extra_context=None):
        """
        View to manage closure of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))
        title = "Close this ePFP"

        if request.method == 'POST':
            url = reverse('admin:prescription_prescription_detail',
                          args=[str(obj.id)])
            if request.POST.get('_cancel'):
                return HttpResponseRedirect(url)

            if obj.is_approved and obj.post_state.post_burn_checklist:
                # close the burn
                obj.status = obj.STATUS_CLOSED
                obj.status_modified = timezone.now()
                obj.save()

                # Remove all object permissions for sections that can't be
                # modified after closure. (Pretty much everything.)
                update_permissions(obj, self.admin_site, 'closure')

                self.message_user(
                    request, "Burn closed successfully.")
                return HttpResponseRedirect(url)

        context = {
            'title': title,
            'current': obj,
        }
        return TemplateResponse(request, "admin/prescription/prescription/"
                                "closure.html", context,
                                current_app=self.admin_site.name)

    def add_objectives(self, request, object_id):
        """
        Custom view to allow the user to select which objectives from the
        prescription's region fit.
        """
        # TODO: make a form to handle this...
        obj = self.get_object(request, unquote(object_id))
        if request.method == "POST":
            selected = request.POST.getlist('objectives')
            added, removed = False, False
            for objective in RegionalObjective.objects.filter(
                pk__in=selected
            ).exclude(pk__in=obj.regional_objectives.all()):
                obj.regional_objectives.add(objective)
                added = True
            for objective in obj.regional_objectives.exclude(pk__in=selected):
                obj.regional_objectives.remove(objective)
                removed = True
            if added and removed:
                message = "added and removed successfully."
            elif added:
                message = "added successfully."
            elif removed:
                message = "removed successfully."
            else:
                message = "were not changed."
            self.message_user(request, "Regional objectives " + message)
            url = request.REQUEST.get('next', reverse(
                'admin:prescription_prescription_detail', args=[str(obj.id)]))
            return HttpResponseRedirect(url)

        objectives = RegionalObjective.objects.filter(
            region=obj.region).exclude(pk__in=obj.regional_objectives.all())
        context = {
            'current': obj,
            'objectives': objectives,
        }
        return TemplateResponse(request, self.objectives_template,
                                context, current_app=self.admin_site.name)

    def summary(self, request, object_id):
        """
        A custom view to display section A1 of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))

        if obj is None:
            raise Http404(_('%(name)s object with primary key %(key)r'
                            ' does not exist.') %
                          {'name': force_text(self.opts.verbose_name),
                           'key': object_id})

        if request.method == "POST":
            form = SummaryCompletionStateForm(request.POST,
                                              instance=obj.pre_state)
            if form.is_valid():
                form.save()
                if request.is_ajax():
                    message = "Summary updated."
                    return HttpResponse(json.dumps({'message': message}))
                else:
                    url = reverse('admin:index')
                    return HttpResponseRedirect(url)
            else:
                if request.is_ajax():
                    return HttpResponse(json.dumps({'errors': form.errors}))
        else:
            form = SummaryCompletionStateForm(instance=obj.pre_state)
        context = {
            'current': obj,
            'form': form,
        }
        return TemplateResponse(request, self.summary_template,
                                context, current_app=self.admin_site.name)

    def pre_summary(self, request, object_id):
        """
        A custom view to display section A1 of an ePFP.
        """
        class AdminPrescriptionSummaryForm(PrescriptionSummaryForm):
            formfield_callback = partial(
                self.formfield_for_dbfield, request=request)

        obj = self.get_object(request, unquote(object_id))
        AdminPrescriptionSummaryForm = self.get_form(request, obj)

        if request.method == "POST":
            form = AdminPrescriptionSummaryForm(request.POST, instance=obj)
            if form.is_valid():
                form.save()
                if request.is_ajax():
                    message = obj.__str__() + ' summary updated successfully.'
                    return HttpResponse(json.dumps({'message': message}))
                else:
                    return HttpResponseRedirect(obj.get_absolute_url())
            else:
                if request.is_ajax():
                    return HttpResponse(json.dumps({'errors': form.errors}))
        else:
            form = AdminPrescriptionSummaryForm(instance=obj)

        admin_form = helpers.AdminForm(
            form, [(None, {'fields': list(form.base_fields)})],
            self.get_prepopulated_fields(request, obj),
            self.get_readonly_fields(request, obj),
            model_admin=self)
        media = self.media + admin_form.media

        context = {
            'current': obj,
            'form': admin_form,
            'media': media,
            'max_risk': obj.get_maximum_risk,
            'max_complexity': obj.get_maximum_complexity,
            'purposes': [p.name for p in obj.purposes.all()]
        }
        return TemplateResponse(request, self.pre_summary_template,
                                context, current_app=self.admin_site.name)

    def day_summary(self, request, object_id):
        """
        A custom view to display the summary of Part B of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))

        if obj is None:
            raise Http404(_('%(name)s object with primary key (%key)r'
                            ' does not exist.') % {
                                'name': force_text(self.opts.verbose_name),
                                'key': object_id})

        if request.method == "POST":
            form = BurnImplementationStateForm(request.POST,
                                               instance=obj.day_state)
            if form.is_valid():
                form.save()
                if request.is_ajax():
                    message = "Summary updated."
                    return HttpResponse(json.dumps({'message': message}))
                else:
                    url = reverse('admin:index')
                    return HttpResponseRedirect(url)
            else:
                if request.is_ajax():
                    return HttpResponse(json.dumps({'errors': form.errors}))
        else:
            form = BurnImplementationStateForm(instance=obj.day_state)

        context = {
            'current': obj,
            'form': form,
        }
        return TemplateResponse(request, self.day_summary_template,
                                context, current_app=self.admin_site.name)

    def post_summary(self, request, object_id):
        """
        A custom view to display the summary of Part C of an ePFP.
        """
        obj = self.get_object(request, unquote(object_id))

        if obj is None:
            raise Http404(_('%(name)s object with primary key (%key)r'
                            ' does not exist.') % {
                                'name': force_text(self.opts.verbose_name),
                                'key': object_id})

        if request.method == "POST":
            form = BurnClosureStateForm(request.POST, instance=obj.post_state)
            if form.is_valid():
                form.save()
                if request.is_ajax():
                    message = "Summary updated."
                    return HttpResponse(json.dumps({'message': message}))
                else:
                    url = reverse('admin:index')
                    return HttpResponseRedirect(url)
            else:
                if request.is_ajax():
                    return HttpResponse(json.dumps({'errors': form.errors}))
        else:
            form = BurnClosureStateForm(instance=obj.post_state)

        context = {
            'current': obj,
            'form': form,
        }
        return TemplateResponse(request, self.post_summary_template,
                                context, current_app=self.admin_site.name)

    def lualatex(self, request, object_id):
        obj = self.get_object(request, unquote(object_id))
        template = request.GET.get("template", "pfp")
        response = HttpResponse(content_type='application/pdf')
        texname = template + ".tex"
        filename = template + ".pdf"
        now = timezone.localtime(timezone.now())
        timestamp = now.isoformat().rsplit(
            ".")[0].replace(":", "")[:-2]
        downloadname = "{0}_{1}_{2}_{3}".format(
            obj.season, obj.burn_id, timestamp, filename).replace(
                ' ', '_')
        subtitles = {
            "parta": "Part A - Summary & Approval",
            "partb": "Part B - Burn Implementation Plan",
            "partc": "Part C - Burn Closure and Evaluation",
            "partd": "Part D - Supporting Documents and Maps"
        }
        context = {
            'current': obj,
            'prescription': obj,
            'embed': request.GET.get("embed", True),
            'headers': request.GET.get("headers", True),
            'title': request.GET.get("title", "Prescribed Fire Plan"),
            'subtitle': subtitles.get(template, ""),
            'timestamp': now,
            'downloadname': downloadname,
            'settings': settings,
            'baseurl': request.build_absolute_uri("/")[:-1]
        }
        if request.GET.get("download", False) is False:
            disposition = "inline"
        else:
            disposition = "attachment"
        response['Content-Disposition'] = (
            '{0}; filename="{1}"'.format(
                disposition, downloadname))
        output = render_to_string(
            "latex/" + template + ".tex", context,
            context_instance=RequestContext(request))
        hgrepo = os.path.join(settings.MEDIA_ROOT, "prescriptions")
        if not os.path.exists(hgrepo):
            os.makedirs(hgrepo)
        if not os.path.exists(os.path.join(hgrepo, ".hg")):
            os.chdir(hgrepo)
            subprocess.check_output(["hg", "init"])

        # directory should be a property of prescription model
        # so caching machinering can put outdated flag in directory
        # to trigger a cache repop next download
        directory = os.path.join(settings.MEDIA_ROOT, 'prescriptions',
                                 str(obj.season), obj.burn_id)
        if not os.path.exists(directory):
            os.makedirs(directory)
        os.chdir(directory)

        if os.path.exists(filename):
            # if outdated then remove all pdf's
            if os.path.exists("outdated"):
                for f in os.listdir("."):
                    if os.path.splitext(f)[0] == ".pdf":
                        os.remove(f)
            # If cache is valid just return
            else:
                # Until we set outdated file don't cache
                os.remove(filename)
                #with open(filename, "r") as f:
                #    response.write(f.read())
                #return response

        with open(texname, "w") as f:
            f.write(output.encode('utf-8'))

        cmd = ['lualatex',
               "--interaction", "batchmode",
               texname]
        try:
            for i in range(2):
                # 2 passes for numbering
                try:
                    subprocess.check_output(cmd)
                except subprocess.CalledProcessError:
                    pass
            subprocess.check_output(["hg", "addr"])
            subprocess.check_output(["hg", "commit", "-m", "updated pdf"])
        except subprocess.CalledProcessError:
            if not os.path.exists(filename):
                filename = filename.replace(".pdf", ".log")
                response["Content-Type"] = "text"
            else:
                raise

        with open(filename, "r") as f:
            response.write(f.read())

        return response


class PrescriptionMixin(object):
    prescription_filter_field = "prescription"

    def __init__(self, model, admin_site):
        self.prescription = None
        return super(PrescriptionMixin, self).__init__(model, admin_site)

    def get_urls(self):
        from django.conf.urls import patterns, url

        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)
            return update_wrapper(wrapper, view)

        info = self.model._meta.app_label, self.model._meta.module_name

        urlpatterns = patterns(
            '',
            url(r'^prescription/(\d+)/$',
                wrap(self.changelist_view),
                name='%s_%s_changelist' % info),
            url(r'^add/prescription/(\d+)/$',
                wrap(self.add_view),
                name='%s_%s_add' % info),
            url(r'^(.+)/history/prescription/(\d+)/$',
                wrap(self.history_view),
                name='%s_%s_history' % info),
            url(r'^(.+)/delete/prescription/(\d+)/$',
                wrap(self.delete_view),
                name='%s_%s_delete' % info),
            url(r'^(.+)/prescription/(\d+)/$',
                wrap(self.change_view),
                name='%s_%s_change' % info)
        )
        return urlpatterns

    def get_changelist(self, request, **kwargs):
        ChangeList = super(PrescriptionMixin, self).get_changelist(
            request, **kwargs)
        admin = self

        class PrescriptionChangeList(ChangeList):
            def get_query_set(self, request):
                qs = super(PrescriptionChangeList, self).get_query_set(request)
                if admin.prescription is not None:
                    fields = {
                        admin.prescription_filter_field: admin.prescription
                    }
                    return qs.filter(**fields)
                else:
                    return qs

            def url_for_result(self, result):
                pk = getattr(result, self.pk_attname)
                return reverse('admin:%s_%s_change' % (self.opts.app_label,
                                                       self.opts.module_name),
                               args=(quote(pk), quote(admin.prescription.pk)),
                               current_app=self.model_admin.admin_site.name)

        return PrescriptionChangeList

    def get_prescription(self, request, object_id):
        """
        Returns an instance matching the primary key provided. ``None`` is
        returned if no match is found (or the object_id failed validation
        against the primary key field).
        WARNING: be very careful here -- setting state on a thread-local
        object so we have to ensure that it changes "per-request" and the
        only reason we set the state is to access it in methods that don't
        include prescription_id. This feels just as hacky as request.session.
        """
        queryset = Prescription.objects
        model = Prescription

        try:
            object_id = model._meta.pk.to_python(object_id)
            self.prescription = queryset.get(pk=object_id)
        except (Prescription.DoesNotExist, ValidationError, ValueError):
            self.prescription = None
        return self.prescription

    def changelist_view(self, request, prescription_id, extra_context=None):
        prescription = self.get_prescription(request, unquote(prescription_id))

        if prescription is None:
            raise Http404(_('prescription object with primary key %(key)r '
                            'does not exist.') % {'key': prescription_id})

        editable = self.get_list_editable(request)
        if not editable or 'id' in editable and len(editable) == 1:
            editable = False
        else:
            editable = True

        context = {
            'current': prescription,
            'editable': editable
        }
        context.update(extra_context or {})

        return super(PrescriptionMixin, self).changelist_view(
            request, extra_context=context)

    def add_view(self, request, prescription_id, form_url='',
                 extra_context=None):
        prescription = self.get_prescription(request, unquote(prescription_id))

        if prescription is None:
            raise Http404(_('prescription object with primary key %(key)r '
                            'does not exist.') % {'key': prescription_id})

        context = {
            'current': prescription
        }
        context.update(extra_context or {})

        return super(PrescriptionMixin, self).add_view(
            request, extra_context=context)

    def change_view(self, request, object_id, prescription_id,
                    extra_context=None):
        prescription = self.get_prescription(request, unquote(prescription_id))

        if prescription is None:
            raise Http404(_('prescription object with primary key %(key)r '
                            'does not exist.') % {'key': prescription_id})

        if request.method == 'POST' and "_saveasnew" in request.POST:
            opts = self.model._meta
            return self.add_view(request, prescription_id=prescription.id,
                                 form_url=reverse(
                                     'admin:%s_%s_add' %
                                     (opts.app_label, opts.module_name),
                                     args=[prescription.id],
                                     current_app=self.admin_site.name))

        context = {
            'current': prescription
        }
        context.update(extra_context or {})

        return super(PrescriptionMixin, self).change_view(
            request, object_id, extra_context=context)

    def history_view(self, request, object_id, prescription_id,
                     extra_context=None):
        prescription = self.get_prescription(request, unquote(prescription_id))

        if prescription is None:
            raise Http404(_('prescription object with primary key %(key)r '
                            'does not exist.') % {'key': prescription_id})

        context = {
            'current': prescription
        }
        context.update(extra_context or {})

        return super(PrescriptionMixin, self).history_view(
            request, object_id, extra_context=context)

    @csrf_protect_m
    @transaction.commit_on_success
    def delete_view(self, request, object_id, prescription_id,
                    extra_context=None):
        "The 'delete' admin view for this model."
        opts = self.model._meta
        app_label = opts.app_label

        obj = self.get_object(request, unquote(object_id))
        prescription = self.get_prescription(request, unquote(prescription_id))

        if not self.has_delete_permission(request, obj):
            raise PermissionDenied

        if obj is None:
            raise Http404(
                _('%(name)s object with primary key %(key)r does '
                  'not exist.') % {'name': force_text(opts.verbose_name),
                                   'key': escape(object_id)})

        if prescription is None:
            raise Http404(
                _('prescription object with primary key %(key)r does not '
                  'exist.') % {'key': prescription_id})

        using = router.db_for_write(self.model)

        # Populate deleted_objects, a data structure of all related objects
        # that will also be deleted.
        (deleted_objects, perms_needed, protected) = get_deleted_objects(
            [obj], opts, request.user, self.admin_site, using)

        if request.POST:    # The user has already confirmed the deletion.
            if perms_needed:
                raise PermissionDenied
            obj_display = force_text(obj)
            self.log_deletion(request, obj, obj_display)
            self.delete_model(request, obj)

            self.message_user(request, _(
                'The %(name)s "%(obj)s" was deleted successfully.') % {
                    'name': force_text(opts.verbose_name),
                    'obj': force_text(obj_display)},
                messages.SUCCESS)

            if self.has_change_permission(request, None):
                if "next" in request.GET:
                    post_url = request.GET['next']
                else:
                    post_url = reverse(
                        'admin:%s_%s_changelist' % (opts.app_label,
                                                    opts.module_name),
                        args=(quote(self.prescription.pk),),
                        current_app=self.admin_site.name)
            else:
                post_url = reverse('admin:index',
                                   current_app=self.admin_site.name)
            return HttpResponseRedirect(post_url)

        object_name = force_text(opts.verbose_name)

        if perms_needed or protected:
            title = _("Cannot delete %(name)s") % {"name": object_name}
        else:
            title = _("Are you sure?")

        context = {
            "title": title,
            "object_name": object_name,
            'current': prescription,
            "object": obj,
            "deleted_objects": deleted_objects,
            "perms_lacking": perms_needed,
            "protected": protected,
            "opts": opts,
            "app_label": app_label,
        }
        context.update(extra_context or {})

        return TemplateResponse(request, self.delete_confirmation_template or [
            "admin/%s/delete_confirmation.html" % app_label,
            "admin/delete_confirmation.html"
        ], context, current_app=self.admin_site.name)

    def get_readonly_fields(self, request, obj=None):
        """
        For any of our ePFP related admins, do not allow editing if the ePFP
        has been submitted for endorsement, has been endorsed, or has been
        approved.
        """
        current = self.prescription
        if not current.is_draft:
            return self.list_editable
        else:
            return super(PrescriptionMixin, self).get_readonly_fields(
                request, obj)

    def get_changelist_form(self, request, **kwargs):
        return self.get_form(request, **kwargs)

    def get_list_display(self, request):
        """
        Allow the use of the current request inside our remove function.
        This lets us check if the current user has permission to delete a
        particular object.
        """
        delete = partial(self.remove, request=request)
        delete.short_description = ""
        delete.allow_tags = True
        list_display = list(self.list_display)
        for index, field_name in enumerate(list_display):
            field = getattr(self.model, field_name, None)
            if hasattr(field, "related"):
                list_display.remove(field_name)
                list_display.insert(
                    index, self.display_add_link(request, field.related))
        list_display.append(delete)
        return list_display

    def remove(self, obj, **kwargs):
        """
        This will not work without a custom get_list_display like above in
        this class.
        """
        request = kwargs.pop('request')
        if self.has_delete_permission(request, obj):
            info = obj._meta.app_label, obj._meta.module_name
            delete_url = reverse('admin:%s_%s_delete' % info,
                                 args=(quote(obj.pk),
                                       quote(self.prescription.pk)))
            return ('<div><a href="%s" class="inline-deletelink"'
                    'title="Delete"></a></div>') % delete_url
        else:
            return ""

    def display_add_link(self, request, related):
        def inner(obj):
            opts = related.model._meta
            kwargs = {related.field.name: obj}
            count = related.model._default_manager.filter(**kwargs).count()
            context = {
                'related': related,
                'obj': obj,
                'opts': opts,
                'count': count
            }
            return render_to_string(
                'admin/change_list_links.html',
                RequestContext(request, context)
            )
        inner.allow_tags = True
        inner.short_description = related.opts.verbose_name_plural.title()
        return inner

    def response_add(self, request, obj, post_url_continue=None):
        redirect = request.GET['next'] if "next" in request.GET else None
        if redirect is None:
            opts = obj._meta
            pk_value = obj._get_pk_val()
            redirect = reverse('admin:%s_%s_change' %
                               (opts.app_label, opts.module_name),
                               args=(pk_value, self.prescription.pk),
                               current_app=self.admin_site.name)

        return super(PrescriptionMixin, self).response_add(
            request, obj, post_url_continue=redirect)

    def response_change(self, request, obj):
        opts = self.model._meta
        pk_value = obj._get_pk_val()
        msg_dict = {'name': force_text(opts.verbose_name),
                    'obj': force_text(obj)}
        if "_saveasnew" in request.POST:
            msg = ('The %(name)s "%(obj)s" was added successfully. You may ' +
                   ' edit it again below.' % msg_dict)
            self.message_user(request, msg)
            return HttpResponseRedirect(reverse('admin:%s_%s_change' %
                                        (opts.app_label, opts.module_name),
                                        args=(pk_value, self.prescription.pk),
                                        current_app=self.admin_site.name))
        elif "_addanother" in request.POST:
            msg = ('The %(name)s "%(obj)s" was changed successfully. You may' +
                   ' add another %(name)s below.' % msg_dict)
            self.message_user(request, msg)
            return HttpResponseRedirect(reverse('admin:%s_%s_add' %
                                        (opts.app_label, opts.module_name),
                                        args=(self.prescription.pk,),
                                        current_app=self.admin_site.name))
        return super(PrescriptionMixin, self).response_change(request, obj)

    def response_post_save_add(self, request, obj):
        """
        Figure out where to redirect after the 'Save' button has been pressed
        when adding a new object.
        """
        opts = self.model._meta

        if "next" in request.GET:
            return HttpResponseRedirect(request.GET['next'])

        if self.has_change_permission(request, None):
            post_url = reverse('admin:%s_%s_changelist' %
                               (opts.app_label, opts.module_name),
                               args=(quote(self.prescription.pk),),
                               current_app=self.admin_site.name)
        else:
            post_url = reverse('admin:index',
                               current_app=self.admin_site.name)

        return HttpResponseRedirect(post_url)

    def response_post_save_change(self, request, obj):
        """
        Figure out where to redirect after the 'Save' button has been pressed
        when editing an existing object.
        """
        opts = self.model._meta

        if "next" in request.GET:
            return HttpResponseRedirect(request.GET['next'])

        if self.has_change_permission(request, None):
            post_url = reverse('admin:%s_%s_changelist' %
                               (opts.app_label, opts.module_name),
                               args=(quote(self.prescription.pk),),
                               current_app=self.admin_site.name)
        else:
            post_url = reverse('admin:index',
                               current_app=self.admin_site.name)

        return HttpResponseRedirect(post_url)


class SavePrescriptionMixin(object):
    exclude = ('prescription',)
    can_delete = False
    lock_after = 'endorsement'

    def get_list_editable(self, request):
        current = self.prescription
        if ((self.lock_after == 'endorsement' and not current.is_draft) or
            (self.lock_after == 'closure' and current.is_closed)):
            return ('id',)
        else:
            return self.list_editable

    def save_model(self, request, obj, form, change):
        """
        Save the model and assign delete permissions to particular objects.
        """
        obj.prescription = self.prescription
        obj.save()

        # If can_delete is set, allow the user to delete this object.
        if self.can_delete:
            opts = self.opts
            group = Group.objects.get(name='Users')
            perm = get_permission_codename('delete', opts)
            assign_perm("%s.%s" % (opts.app_label, perm), group, obj)


class ObjectiveAdmin(PrescriptionMixin, SavePrescriptionMixin,
                     BaseAdmin):
    list_display = ("objectives",)
    list_editable = ("objectives",)
    list_empty_form = True
    list_display_links = (None,)
    actions = None
    can_delete = True
    lock_after = "endorsement"


class RegionalObjectiveAdmin(admin.ModelAdmin):
    list_display = ("region", "impact", "fma_names", "objectives",)
    list_filter = ("region", "impact")

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """
        Override to default the region to the user's profile region.
        """
        if db_field.name == 'region':
            if request.user.profile.region is not None:
                kwargs['initial'] = request.user.profile.region.pk
            return db_field.formfield(**kwargs)

        return super(PrescriptionAdmin, self).formfield_for_foreignkey(
            db_field, request, **kwargs)


class SuccessCriteriaAdmin(PrescriptionMixin, SavePrescriptionMixin,
                           BaseAdmin):
    list_display = ("criteria",)
    list_editable = ("criteria",)
    list_empty_form = True
    list_display_links = (None,)
    actions = None
    can_delete = True
    lock_after = 'endorsement'


class PriorityJustificationAdmin(PrescriptionMixin, SavePrescriptionMixin,
                                 BaseAdmin):
    list_display = ("purpose", "criteria_display", "rationale", "priority")
    list_editable = ("rationale", "priority")
    list_display_links = (None,)
    actions = None

    def criteria_display(self, obj):
        """
        Fix up the display of the criteria so that it looks a bit nicer.
        """
        return markdownify(obj.criteria)
    criteria_display.short_description = "Criteria"

    def queryset(self, request):
        qs = super(PriorityJustificationAdmin, self).queryset(request)
        return qs.filter(relevant=True)

    def get_changelist_formset(self, request):
        FormSet = super(PriorityJustificationAdmin,
                        self).get_changelist_formset(request)

        current = self.prescription

        # we are patching the formset here to include an extra form
        # note: this form must be popped before it is processed by the
        # result_list (result_list needs a set of forms of the same type)
        class PriorityFormSet(FormSet):
            def _construct_forms(self):
                super(PriorityFormSet, self)._construct_forms()
                self.forms.append(PrescriptionPriorityForm(
                    request.POST or None, instance=current))

            def full_clean(self):
                super(PriorityFormSet, self).full_clean()
                if self.is_bound:
                    special_form = self.forms[self.total_form_count()]
                    self._errors.append(special_form.errors)

            def is_valid(self):
                # run .is_valid() on our "special" last form
                forms_valid = super(PriorityFormSet, self).is_valid()
                return (forms_valid and
                        self.forms[self.total_form_count()].is_valid())
        return PriorityFormSet

    def changelist_view(self, request, prescription_id, extra_context=None):
        current = self.get_prescription(request, unquote(prescription_id))

        if request.method == "POST" and current.is_draft:
            # we are sort of constructing this form twice (once in the
            # changelist and second time here) but whatever :)
            # we save this form here if it has changed, so the subsequent call
            # to form.has_changed() in the admin.changelist_view() should
            # return False, we don't want admin.changelist_view() to mess up
            # with this "special" form
            form = PrescriptionPriorityForm(request.POST or None,
                                            instance=current)
            # note: the .is_valid() updates the instance with the data from
            # the form
            current_priority = current.priority
            current_rationale = current.rationale
            if form.is_valid() and form.has_changed():
                if (str(form.data['priority']) != str(current_priority)
                        and form.data['rationale'] != current_rationale):
                    message = "Saved overall priority and rationale."
                elif str(form.data['priority']) != str(current_priority):
                    message = "Saved overall priority."
                else:
                    message = "Saved overall rationale."
                form.save()
                self.message_user(request, message)

        return super(PriorityJustificationAdmin, self).changelist_view(
            request, prescription_id, extra_context=extra_context)


class BriefingChecklistAdmin(PrescriptionMixin, SavePrescriptionMixin,
                             BaseAdmin):
    list_display = ("title_column", "notes")
    list_group_by = "smeac_category"
    list_editable = ("notes",)
    actions = None
    form = BriefingChecklistForm
    list_display_links = (None,)

    def get_action(self, obj):
        if obj.action:
            return obj.action.description
        else:
            return ""
    get_action.short_description = "Action"

    def smeac_category(self, obj):
        return obj.smeac.category

    def title_column(self, obj):
        return obj.title
    title_column.short_description = "Topic"

    def get_readonly_fields(self, request, obj=None):
        if self.prescription.is_closed:
            return self.list_editable
        else:
            return self.readonly_fields
